@@ -1,10 +1,10 @@
 /*  Plzip - Parallel compressor compatible with lzip
     Copyright (C) 2009 Laszlo Ersek.
-    Copyright (C) 2009, 2010, 2011, 2012, 2013 Antonio Diaz Diaz.
+    Copyright (C) 2009-2015 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
+    the Free Software Foundation, either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
@@ -47,6 +47,8 @@ struct Packet			// data block
   {
   uint8_t * data;		// data == 0 means end of member
   int size;			// number of bytes in data (if any)
+  explicit Packet( uint8_t * const d = 0, const int s = 0 )
+    : data( d ), size( s ) {}
   };
 
 
@@ -65,35 +67,36 @@ private:
   std::vector< std::queue< Packet * > > opacket_queues;
   int num_working;			// number of workers still running
   const int num_workers;		// number of workers
-  const int num_slots;			// max output packets in circulation
-  int num_free;				// remaining free output slots
+  const unsigned out_slots;		// max output packets per queue
   pthread_mutex_t imutex;
   pthread_cond_t iav_or_eof;	// input packet available or splitter done
   pthread_mutex_t omutex;
   pthread_cond_t oav_or_exit;	// output packet available or all workers exited
-  pthread_cond_t slot_av;		// free output slot available
-  bool eof;				// splitter done
+  std::vector< pthread_cond_t > slot_av;	// output slot available
+  bool eof;					// splitter done
 
   Packet_courier( const Packet_courier & );	// declared as private
   void operator=( const Packet_courier & );	// declared as private
 
 public:
-  Packet_courier( const int workers, const int slots )
+  Packet_courier( const int workers, const int in_slots, const int oslots )
     : icheck_counter( 0 ), iwait_counter( 0 ),
       ocheck_counter( 0 ), owait_counter( 0 ),
       receive_worker_id( 0 ), deliver_worker_id( 0 ),
-      slot_tally( slots ), ipacket_queues( workers ),
+      slot_tally( in_slots ), ipacket_queues( workers ),
       opacket_queues( workers ), num_working( workers ),
-      num_workers( workers ), num_slots( 8 * slots ), num_free( num_slots ),
+      num_workers( workers ), out_slots( oslots ), slot_av( workers ),
       eof( false )
     {
     xinit( &imutex ); xinit( &iav_or_eof );
-    xinit( &omutex ); xinit( &oav_or_exit ); xinit( &slot_av );
+    xinit( &omutex ); xinit( &oav_or_exit );
+    for( unsigned i = 0; i < slot_av.size(); ++i ) xinit( &slot_av[i] );
     }
 
   ~Packet_courier()
     {
-    xdestroy( &slot_av ); xdestroy( &oav_or_exit ); xdestroy( &omutex );
+    for( unsigned i = 0; i < slot_av.size(); ++i ) xdestroy( &slot_av[i] );
+    xdestroy( &oav_or_exit ); xdestroy( &omutex );
     xdestroy( &iav_or_eof ); xdestroy( &imutex );
     }
 
@@ -101,9 +104,7 @@ public:
   // if data == 0, move to next queue
   void receive_packet( uint8_t * const data, const int size )
     {
-    Packet * ipacket = new Packet;
-    ipacket->data = data;
-    ipacket->size = size;
+    Packet * const ipacket = new Packet( data, size );
     if( data )
       { in_size += size; slot_tally.get_slot(); }  // wait for a free slot
     xlock( &imutex );
@@ -149,9 +150,8 @@ public:
     xlock( &omutex );
     if( opacket->data )
       {
-      while( worker_id != deliver_worker_id && num_free <= 0 )
-        xwait( &slot_av, &omutex );
-      --num_free;
+      while( opacket_queues[worker_id].size() >= out_slots )
+        xwait( &slot_av[worker_id], &omutex );
       }
     opacket_queues[worker_id].push( opacket );
     if( worker_id == deliver_worker_id ) xsignal( &oav_or_exit );
@@ -175,19 +175,23 @@ public:
       if( opacket_queues[deliver_worker_id].empty() ) break;
       opacket = opacket_queues[deliver_worker_id].front();
       opacket_queues[deliver_worker_id].pop();
-      if( opacket->data )
-        {
-        if( ++num_free == 1 ) xsignal( &slot_av );
-        break;
-        }
+      if( opacket_queues[deliver_worker_id].size() + 1 == out_slots )
+        xsignal( &slot_av[deliver_worker_id] );
+      if( opacket->data ) break;
       if( ++deliver_worker_id >= num_workers ) deliver_worker_id = 0;
-      xbroadcast( &slot_av );		// restart deliver_worker_id thread
       delete opacket; opacket = 0;
       }
     xunlock( &omutex );
     return opacket;
     }
 
+  void add_out_size( const unsigned long long partial_out_size )
+    {
+    xlock( &omutex );
+    out_size += partial_out_size;
+    xunlock( &omutex );
+    }
+
   void finish()			// splitter has no more packets to send
     {
     xlock( &imutex );
@@ -198,8 +202,7 @@ public:
 
   bool finished()		// all packets delivered to muxer
     {
-    if( !slot_tally.all_free() ||
-        num_free != num_slots || !eof || num_working != 0 ) return false;
+    if( !slot_tally.all_free() || !eof || num_working != 0 ) return false;
     for( int i = 0; i < num_workers; ++i )
       if( !ipacket_queues[i].empty() ) return false;
     for( int i = 0; i < num_workers; ++i )
@@ -210,7 +213,7 @@ public:
 
 
 // Search forward from 'pos' for "LZIP" (Boyer-Moore algorithm)
-// Return pos of found string or 'pos+size' if not found.
+// Returns pos of found string or 'pos+size' if not found.
 //
 int find_magic( const uint8_t * const buffer, const int pos, const int size )
   {
@@ -253,7 +256,7 @@ extern "C" void * dsplitter_s( void * arg )
   const int buffer_size = max_packet_size;
   const int base_buffer_size = tsize + buffer_size + hsize;
   uint8_t * const base_buffer = new( std::nothrow ) uint8_t[base_buffer_size];
-  if( !base_buffer ) { pp( "Not enough memory" ); cleanup_and_fail(); }
+  if( !base_buffer ) { pp( "Not enough memory." ); cleanup_and_fail(); }
   uint8_t * const buffer = base_buffer + tsize;
 
   int size = readblock( infd, buffer, buffer_size + hsize ) - hsize;
@@ -261,10 +264,10 @@ extern "C" void * dsplitter_s( void * arg )
   if( size != buffer_size && errno )
     { pp(); show_error( "Read error", errno ); cleanup_and_fail(); }
   if( size + hsize < min_member_size )
-    { pp( "Input file is too short" ); cleanup_and_fail( 2 ); }
+    { pp( "Input file is too short." ); cleanup_and_fail( 2 ); }
   const File_header & header = *(File_header *)buffer;
   if( !header.verify_magic() )
-    { pp( "Bad magic number (file not in lzip format)" ); cleanup_and_fail( 2 ); }
+    { pp( "Bad magic number (file not in lzip format)." ); cleanup_and_fail( 2 ); }
   if( !header.verify_version() )
     {
     if( verbosity >= 0 )
@@ -273,6 +276,7 @@ extern "C" void * dsplitter_s( void * arg )
                       header.version() ); }
     cleanup_and_fail( 2 );
     }
+  show_header( header.dictionary_size() );
 
   unsigned long long partial_member_size = 0;
   while( true )
@@ -297,7 +301,7 @@ extern "C" void * dsplitter_s( void * arg )
             cleanup_and_fail( 2 );
             }
           uint8_t * const data = new( std::nothrow ) uint8_t[newpos - pos];
-          if( !data ) { pp( "Not enough memory" ); cleanup_and_fail(); }
+          if( !data ) { pp( "Not enough memory." ); cleanup_and_fail(); }
           std::memcpy( data, buffer + pos, newpos - pos );
           courier.receive_packet( data, newpos - pos );
           courier.receive_packet( 0, 0 );	// end of member token
@@ -310,7 +314,7 @@ extern "C" void * dsplitter_s( void * arg )
     if( at_stream_end )
       {
       uint8_t * data = new( std::nothrow ) uint8_t[size + hsize - pos];
-      if( !data ) { pp( "Not enough memory" ); cleanup_and_fail(); }
+      if( !data ) { pp( "Not enough memory." ); cleanup_and_fail(); }
       std::memcpy( data, buffer + pos, size + hsize - pos );
       courier.receive_packet( data, size + hsize - pos );
       courier.receive_packet( 0, 0 );		// end of member token
@@ -320,7 +324,7 @@ extern "C" void * dsplitter_s( void * arg )
       {
       partial_member_size += buffer_size - pos;
       uint8_t * data = new( std::nothrow ) uint8_t[buffer_size - pos];
-      if( !data ) { pp( "Not enough memory" ); cleanup_and_fail(); }
+      if( !data ) { pp( "Not enough memory." ); cleanup_and_fail(); }
       std::memcpy( data, buffer + pos, buffer_size - pos );
       courier.receive_packet( data, buffer_size - pos );
       }
@@ -341,22 +345,25 @@ struct Worker_arg
   Packet_courier * courier;
   const Pretty_print * pp;
   int worker_id;
+  bool testing;
   };
 
 
-       // consume packets from courier, decompress their contents, and
-       // give the produced packets to courier.
+       // consume packets from courier, decompress their contents and,
+       // if not testing, give the produced packets to courier.
 extern "C" void * dworker_s( void * arg )
   {
   const Worker_arg & tmp = *(Worker_arg *)arg;
   Packet_courier & courier = *tmp.courier;
   const Pretty_print & pp = *tmp.pp;
   const int worker_id = tmp.worker_id;
+  const bool testing = tmp.testing;
 
   uint8_t * new_data = new( std::nothrow ) uint8_t[max_packet_size];
   LZ_Decoder * const decoder = LZ_decompress_open();
   if( !new_data || !decoder || LZ_decompress_errno( decoder ) != LZ_ok )
-    { pp( "Not enough memory" ); cleanup_and_fail(); }
+    { pp( "Not enough memory." ); cleanup_and_fail(); }
+  unsigned long long partial_out_size = 0;
   int new_pos = 0;
   bool trailing_garbage_found = false;
 
@@ -373,10 +380,10 @@ extern "C" void * dworker_s( void * arg )
         {
         const int wr = LZ_decompress_write( decoder, ipacket->data + written,
                                             ipacket->size - written );
-        if( wr < 0 ) internal_error( "library error (LZ_decompress_write)" );
+        if( wr < 0 ) internal_error( "library error (LZ_decompress_write)." );
         written += wr;
         if( written > ipacket->size )
-          internal_error( "ipacket size exceeded in worker" );
+          internal_error( "ipacket size exceeded in worker." );
         }
       while( !trailing_garbage_found )	// read and pack decompressed data
         {
@@ -391,28 +398,25 @@ extern "C" void * dworker_s( void * arg )
           }
         else new_pos += rd;
         if( new_pos > max_packet_size )
-          internal_error( "opacket size exceeded in worker" );
+          internal_error( "opacket size exceeded in worker." );
         if( new_pos == max_packet_size || trailing_garbage_found ||
             LZ_decompress_finished( decoder ) == 1 )
           {
-          if( new_pos > 0 )			// make data packet
+          if( !testing && new_pos > 0 )		// make data packet
             {
-            Packet * opacket = new Packet;
-            opacket->data = new_data;
-            opacket->size = new_pos;
+            Packet * const opacket = new Packet( new_data, new_pos );
             courier.collect_packet( opacket, worker_id );
-            new_pos = 0;
             new_data = new( std::nothrow ) uint8_t[max_packet_size];
-            if( !new_data ) { pp( "Not enough memory" ); cleanup_and_fail(); }
+            if( !new_data ) { pp( "Not enough memory." ); cleanup_and_fail(); }
             }
+          partial_out_size += new_pos;
+          new_pos = 0;
           if( trailing_garbage_found ||
               LZ_decompress_finished( decoder ) == 1 )
             {
-            LZ_decompress_reset( decoder );	// prepare for new ipacket
-            Packet * opacket = new Packet;	// end of member token
-            opacket->data = 0;
-            opacket->size = 0;
-            courier.collect_packet( opacket, worker_id );
+            if( !testing )			// end of member token
+              courier.collect_packet( new Packet, worker_id );
+            LZ_decompress_reset( decoder );	// prepare for new member
             break;
             }
           }
@@ -425,10 +429,11 @@ extern "C" void * dworker_s( void * arg )
     }
 
   delete[] new_data;
+  courier.add_out_size( partial_out_size );
   if( LZ_decompress_member_position( decoder ) != 0 )
-    { pp( "Error, some data remains in decoder" ); cleanup_and_fail(); }
+    { pp( "Error, some data remains in decoder." ); cleanup_and_fail(); }
   if( LZ_decompress_close( decoder ) < 0 )
-    { pp( "LZ_decompress_close failed" ); cleanup_and_fail(); }
+    { pp( "LZ_decompress_close failed." ); cleanup_and_fail(); }
   return 0;
   }
 
@@ -439,17 +444,12 @@ void muxer( Packet_courier & courier, const Pretty_print & pp, const int outfd )
   {
   while( true )
     {
-    Packet * opacket = courier.deliver_packet();
+    Packet * const opacket = courier.deliver_packet();
     if( !opacket ) break;	// queue is empty. all workers exited
 
-    out_size += opacket->size;
-
-    if( outfd >= 0 )
-      {
-      const int wr = writeblock( outfd, opacket->data, opacket->size );
-      if( wr != opacket->size )
-        { pp(); show_error( "Write error", errno ); cleanup_and_fail(); }
-      }
+    const int wr = writeblock( outfd, opacket->data, opacket->size );
+    if( wr != opacket->size )
+      { pp(); show_error( "Write error", errno ); cleanup_and_fail(); }
     delete[] opacket->data;
     delete opacket;
     }
@@ -458,18 +458,18 @@ void muxer( Packet_courier & courier, const Pretty_print & pp, const int outfd )
 } // end namespace
 
 
-    // init the courier, then start the splitter and the workers and
-    // call the muxer.
+    // init the courier, then start the splitter and the workers and,
+    // if not testing, call the muxer.
 int dec_stream( const int num_workers, const int infd, const int outfd,
-                const Pretty_print & pp, const int debug_level,
-                const bool testing )
+                const Pretty_print & pp, const int debug_level )
   {
-  const int slots_per_worker = 2;
-  const int num_slots = ( ( INT_MAX / num_workers >= slots_per_worker ) ?
-                          num_workers * slots_per_worker : INT_MAX );
+  const int in_slots_per_worker = 2;
+  const int out_slots = 32;
+  const int in_slots = ( INT_MAX / num_workers >= in_slots_per_worker ) ?
+                       num_workers * in_slots_per_worker : INT_MAX;
   in_size = 0;
   out_size = 0;
-  Packet_courier courier( num_workers, num_slots );
+  Packet_courier courier( num_workers, in_slots, out_slots );
 
   Splitter_arg splitter_arg;
   splitter_arg.courier = &courier;
@@ -484,18 +484,19 @@ int dec_stream( const int num_workers, const int infd, const int outfd,
   Worker_arg * worker_args = new( std::nothrow ) Worker_arg[num_workers];
   pthread_t * worker_threads = new( std::nothrow ) pthread_t[num_workers];
   if( !worker_args || !worker_threads )
-    { pp( "Not enough memory" ); cleanup_and_fail(); }
+    { pp( "Not enough memory." ); cleanup_and_fail(); }
   for( int i = 0; i < num_workers; ++i )
     {
     worker_args[i].courier = &courier;
     worker_args[i].pp = &pp;
     worker_args[i].worker_id = i;
+    worker_args[i].testing = ( outfd < 0 );
     errcode = pthread_create( &worker_threads[i], 0, dworker_s, &worker_args[i] );
     if( errcode )
       { show_error( "Can't create worker threads", errcode ); cleanup_and_fail(); }
     }
 
-  muxer( courier, pp, outfd );
+  if( outfd >= 0 ) muxer( courier, pp, outfd );
 
   for( int i = num_workers - 1; i >= 0; --i )
     {
@@ -515,11 +516,11 @@ int dec_stream( const int num_workers, const int infd, const int outfd,
                   (double)out_size / in_size,
                   ( 8.0 * in_size ) / out_size,
                   100.0 * ( 1.0 - ( (double)in_size / out_size ) ) );
-  if( verbosity >= 3 )
+  if( verbosity >= 4 )
     std::fprintf( stderr, "decompressed size %9llu, size %9llu.  ",
                   out_size, in_size );
 
-  if( verbosity >= 1 ) std::fprintf( stderr, testing ? "ok\n" : "done\n" );
+  if( verbosity >= 1 ) std::fprintf( stderr, (outfd < 0) ? "ok\n" : "done\n" );
 
   if( debug_level & 1 )
     std::fprintf( stderr,
@@ -532,6 +533,6 @@ int dec_stream( const int num_workers, const int infd, const int outfd,
       courier.ocheck_counter,
       courier.owait_counter );
 
-  if( !courier.finished() ) internal_error( "courier not finished" );
+  if( !courier.finished() ) internal_error( "courier not finished." );
   return 0;
   }
